Require Import GQM.Base.

Inductive form :=
| Var (x : var)
| Imp (f g : form)
| Box (f : form)
| SquareAll (f : {bind form}).

Instance Ids_form : Ids form. derive. Defined.
Instance Rename_form : Rename form. derive. Defined.
Instance Subst_form : Subst form. derive. Defined.
Instance SubstLemmas_form : SubstLemmas form. derive. Qed.

(* the implicit parameters cause trouble sometimes, because they cannot be infered in all contexts *)
Notation subst_ := (@subst _ Subst_form) (only parsing).
Notation ren_ := (@ren _ Ids_form) (only parsing).
Notation up_ := (@up _ Ids_form Rename_form) (only parsing).
Notation ids_ := (@ids _ Ids_form) (only parsing).

(* see https://github.com/uds-psl/autosubst/blob/1c3bb3bbf5477e3b33533a0fc090399f45fe3034/theories/Autosubst_Classes.v#L54 *)
Notation "s ._[ sigma ]" := (subst_ sigma s)
  (at level 2, sigma at level 200, left associativity,
   format "s ._[ sigma ]", only parsing) : subst_scope.
Notation "s ._[ t /]" := (subst_ (t .: ids_) s)
  (at level 2, t at level 200, left associativity,
   format "s ._[ t /]", only parsing) : subst_scope.
Notation "s ._[ t1 , t2 , .. , tn /]" :=
  (subst_ (scons t1 (scons t2 .. (scons tn ids_) .. )) s)
  (at level 2, left associativity,
format "s '[ ' ._[ t1 , '/' t2 , '/' .. , '/' tn /] ']'", only parsing) : subst_scope.

Notation "'[A]' B"   := (SquareAll B) (at level 74, right associativity).
Notation "|[]| A"   := (Box A) (at level 74, right associativity).
Notation "A '->>' B" := (Imp A B) (at level 77, right associativity).

Definition Bot := [A] (Var 0).
Definition Neg A := A ->> Bot.
Notation "'-!' A" := (Neg A) (at level 73, right associativity).
Definition Top := -! Bot.

Definition Or A B := (-! A ->> B).
Definition And A B := -! ( A ->> -! B).

Notation "A '|||' B" := (Or A B) (at level 76, right associativity).
Infix "&" :=  And (at level 75, right associativity).
Definition Equiv A B := ((A ->> B) & (B ->> A)).
Notation "A '<<->>' B" := (Equiv A B) (at level 78).

Definition All A := [A](A.[ren(+1)]).
Notation "'|A|' B" := (All B) (at level 74, right associativity).
Definition Exists A := -!|A|-!A.
Notation "'|E|' B" := (Exists B) (at level 74, right associativity).

Definition DiamondAll A := [A]|E|A.
Notation "'<A>' B" := (DiamondAll B) (at level 74, right associativity).
Definition DiamondExists A := -![A]-!A.
Notation "'<E>' B" := (DiamondExists B) (at level 74, right associativity).
Definition SquareExists A := <E>|A|A.
Notation "'[E]' B" := (SquareExists B) (at level 74, right associativity).

Inductive Quantifier : Set :=
| QSquareAll : Quantifier
| QDiamondAll : Quantifier
| QSquareExists : Quantifier
| QDiamondExists : Quantifier.

Definition quant_to_form(q : Quantifier) : form -> form :=
  match(q) with
  | QSquareAll => SquareAll
  | QDiamondAll => DiamondAll
  | QSquareExists => SquareExists
  | QDiamondExists => DiamondExists
  end.

Definition quant_to_neg_form(q : Quantifier) : form -> form :=
  match(q) with
  | QSquareAll => DiamondExists
  | QDiamondAll => SquareExists
  | QSquareExists => DiamondAll
  | QDiamondExists => SquareAll
  end.

(** ** Decidable Equality *)

Lemma dceq_v : forall n n0, decidable (Var n = Var n0).
Proof.
  elim.
  - case; by [ left | right].
  - move => n IH.
    case; try by right.
    move => n0.
    case: (IH n0).
    + case => ->.
        by left.
    + move => Hneq.
      right.
      move => [Heq].
      apply: Hneq.
      by rewrite Heq.
Qed.

Lemma dceq_f: forall (A B: form), decidable (A = B).
Proof.
  elim; intros; case: B; try by [right];
    try by [ move => f2; case: (X f2); try by [move => ->; left];
    move => Hneq;
    right;
     move => [Heq]].
  - apply: dceq_v.
  - move => f2 g2.
    case: (X f2);case: (X0 g2);move => H1 H2;subst; try by [left]; right; by move => [].
Qed.

(** ** Substitution *)

Fixpoint max_var(f : form) : var :=
  match f with
  | Var n => S n
  | SquareAll g => pred (max_var g)
  | Imp g h => max (max_var g) (max_var h)
  | Box g => max_var g
  end.

Definition switch_at(n: nat)(f: var -> form)(g: var -> form)(v: var) : form :=
  if v <? n then f v else g v.

Lemma subst_end : forall A f g, A.[f] = A.[switch_at (max_var A) f g].
Proof.
  elim; asimpl.
  - rewrite /switch_at.
    move => x f.
    suff ->: (x <? S x = true) by [].
    apply/PeanoNat.Nat.ltb_spec0.
    omega.
  - move => A IHA B IHB f g.
    rewrite (IHA _ (fun v => if v <? (max (max_var A) (max_var B)) then f v else g v)) (IHB _ (fun v => if v <? (max (max_var A) (max_var B)) then f v else g v)).
    suff ->:(switch_at (max_var A) f (fun v : var => if v <? max (max_var A) (max_var B) then f v else g v) = switch_at (max (max_var A) (max_var B)) f g). {
      suff ->:(switch_at (max_var B) f (fun v : var => if v <? max (max_var A) (max_var B) then f v else g v) = switch_at (max (max_var A) (max_var B)) f g) by [].
      apply: functional_extensionality => n.
      rewrite /switch_at.
      case H: (n <? max (max_var A) (max_var B)); case H2: (n <? (max_var B)) => //.
      exfalso.
      move/negP in H.
      move/idP in H2.
      move/PeanoNat.Nat.ltb_spec0 in H2.
      apply: H.
      apply/PeanoNat.Nat.ltb_spec0.
      apply: Nat.lt_le_trans; first by apply H2.
        by apply: Nat.le_max_r.
    }
    apply: functional_extensionality => n.
    rewrite /switch_at.
    case H: (n <? max (max_var A) (max_var B)); case H2: (n <? (max_var A)) => //.
    exfalso.
    move/negP in H.
    move/idP in H2.
    move/PeanoNat.Nat.ltb_spec0 in H2.
    apply: H.
    apply/PeanoNat.Nat.ltb_spec0.
    apply: Nat.lt_le_trans; first by apply H2.
      by apply: Nat.le_max_l.
  - move => A IH f g.
      by rewrite (IH _ g).
  - move => A IH f g.
    (* Set Printing All. *)
    rewrite (IH _ (up g)).
    suff -> : ((switch_at (max_var A) (up_ f) (up_ g)) = (up_ (switch_at (pred (max_var A)) f g))) by [].
    apply: functional_extensionality => n.
    rewrite /switch_at /up.
    case: n => /=; first by case: (0 <? max_var A).
    move => n.
    case Hn: (S n <? max_var A); case Hn2: (n <? pred (max_var A)) => //.
    + exfalso.
      move/negP in Hn2.
      move/idP in Hn.
      apply: Hn2.
      apply/PeanoNat.Nat.ltb_spec0.
      move/PeanoNat.Nat.ltb_spec0 in Hn.
      omega.
    + exfalso.
      move/negP in Hn.
      move/idP in Hn2.
      apply: Hn.
      apply/PeanoNat.Nat.ltb_spec0.
      move/PeanoNat.Nat.ltb_spec0 in Hn2.
      omega.
Qed.

Fixpoint free_in(v: var)(f: form) : bool :=
  match f with
  | Var n => n =? v
  | Imp A B => free_in v A || free_in v B
  | Box A => free_in v A
  | SquareAll A => free_in (S v) A
  end.

Lemma ids_0_bot : ([A] ids 0) = Bot.
Proof.
    by [].
Qed.

Lemma ids_0_neg : forall A, (A ->> ([A] ids 0)) = -! A.
Proof.
    by [].
Qed.