(** * Different Tactics to make working with GQM easier *)

Require Import GQM.Base.
Require Import GQM.GQM.
Require Import GQM.GQMDeduction.
Require Import GQM.GQMDedSubst.
Require Import GQM.GQMClassical.

(** ** hSplit
 inspired from iSplit in https://gitlab.mpi-sws.org/FP/iris-coq/blob/master/theories/proofmode/tactics.v#L739
 *)

Lemma tac_and_split : forall A B, |-- A -> |-- B -> |-- A & B.
Proof.
  move => A B HA HB.
  apply: gqm_mp.
  - by apply: gqm_mp; first by apply: gqm_andI.
  - done.
Qed.

Tactic Notation "hSplit" := move => *; apply: tac_and_split.

(** ** hAndDestruct *)

Lemma tac_and_destruct_lemma : forall A B, |-- A & B -> |-- A /\ |-- B.
Proof.
  move => A B.
  split.
  - by apply: gqm_mp; first by apply: (gqm_andE1 A B).
  - by apply: gqm_mp; first by apply: (gqm_andE2 A B).
Qed.

Tactic Notation "hAndDestruct" constr(H) := case: (tac_and_destruct_lemma _ _ H); try clear H.

(** ** hRewrite *)

Lemma tac_rewrite_lem_counter : exists A B r, |-- A <<->> B /\ ~ (|-- (r A) <<->> (r B)).
Proof.
  exists (Var 0). exists (-! -! (Var 0)).
  exists (fun f => match f with
           | Var n => Bot
           | _ => -! Bot
           end).
  split.
  - hSplit.
    + apply: gqm_dnegI.
    + apply: gqm_dnegE.
  - move => /= H.
    hAndDestruct H => _ H.
    have H2 := (gqm_mp _ _ H gqm_topI); clear H.
    (* cannot be shown, since |-- is undecidable, but should hold *)
Abort.

(* Set Typeclasses Debug. *)

Class Rewriteable (X : form -> form) := { form_rewrite : forall A B, |-- A <<->> B -> |-- X A <<->> X B }.

Instance id_rewriteable : Rewriteable (fun X => X) := {}. done. Qed.

Instance square_all_rewriteable(r : form -> form) `{Rewriteable r} : Rewriteable (fun X => [A] (r X)) := {}.
Proof.
  case: H => Hr.
  move => A B Heq.
  hAndDestruct (Hr _ _ Heq) => HAB HBA.
  hSplit.
  - apply: gqm_mp; first by apply: gqm_ADist.
    by apply: gqm_Anec.
  - apply: gqm_mp; first by apply: gqm_ADist.
    by apply: gqm_Anec.
Qed.

Instance square_all_id_rewriteable: Rewriteable SquareAll | 100 := {}.
Proof.
  by case: (square_all_rewriteable ssrfun.id).
Qed.

Instance imp_l_rewriteable(r : form -> form) `{Rewriteable r} (X :form) : Rewriteable (fun B => r B ->> X) := {}.
Proof.
  case: H => Hr.
  move => A B Heq.
  hAndDestruct (Hr _ _ Heq) => HAB HBA.
  by hSplit; apply: gqm_mp; try by apply: gqm_chain2.
Qed.

Instance imp_r_rewriteable(r : form -> form) `{Rewriteable r} (X :form) : Rewriteable (fun B => X ->> r B) := {}.
Proof.
  case: H => Hr.
  move => A B Heq.
  hAndDestruct (Hr _ _ Heq) => HAB HBA.
  by hSplit; apply: gqm_mp; try by apply: gqm_chain.
Qed.

Instance imp_rewriteable_id(X :form) : Rewriteable (Imp X) | 100 := {}.
Proof.
  by case: (@imp_r_rewriteable ssrfun.id _ X).
Qed.

Instance imp_both_rewriteable(r1 : form -> form)(r2 : form -> form) `{Rewriteable r1} `{Rewriteable r2} : Rewriteable (fun B => r1 B ->> r2 B) := {}.
Proof.
  case: H => Hr1.
  case: H0 => Hr2.
  move => A B Heq.
  hAndDestruct (Hr1 _ _ Heq) => HAB1 HBA1.
  hAndDestruct (Hr2 _ _ Heq) => HAB2 HBA2.
  hSplit.
  - apply: gqm_mp; last by apply: HBA1.
    apply: gqm_mp; first by apply: gqm_swap.
    apply: gqm_mp; last by apply: HAB2.
    apply: gqm_3chain.
  - apply: gqm_mp; last by apply: HAB1.
    apply: gqm_mp; first by apply: gqm_swap.
    apply: gqm_mp; last by apply: HBA2.
    apply: gqm_3chain.
Qed.


Instance box_rewriteable(r : form -> form) `{Rewriteable r} : Rewriteable (fun X => |[]| (r X)) := {}.
Proof.
  case: H => Hr.
  move => A B Heq.
  have ? := (Hr _ _ Heq).
  apply: gqm_mp; first by apply: gqm_BoxCon.
  by apply: gqm_Anec.
Qed.

Instance box_id_rewriteable: Rewriteable Box | 100 := {}.
Proof.
  by case: (box_rewriteable ssrfun.id).
Qed.

Instance subst_rewriteable(r : form -> form) `{Rewriteable r} (f: var -> form) : Rewriteable (fun X => subst f (r X)) := {}.
Proof.
  case: H => Hr.
  move => A B Heq.
  have := (Hr _ _ Heq); clear Hr Heq.
  have -> : (((r A).[f] <<->> (r B).[f]) = subst f (r A <<->> r B)) by asimpl.
  by apply: gqm_subst.
Qed.

Instance subst_id_rewriteable (f: var -> form): Rewriteable (subst f) | 100 := {}.
Proof.
  by case: (@subst_rewriteable ssrfun.id _ f).
Qed.


(* this is here for testing purposes *)
Lemma gqm_dnegEquiv : forall A, |-- (A <<->> -! -! A).
Proof.
  move => A.
  hSplit.
  - apply: gqm_dnegI.
  - apply: gqm_dnegE.
Qed.

Local Example rewrite_test_1 : forall A B, |-- A ->> B -> |-- A ->> -!-!B.
Proof.
  move => A B Himp.
  hAndDestruct (@form_rewrite (fun X => A ->> X) _ _ _ (gqm_dnegEquiv B)) => Ha _.
  apply: gqm_mp; first by apply: Ha.
  done.
Qed.

Local Example rewrite_test_2 : forall A B, |-- [A](A ->> B) -> |-- [A](A ->> -!-!B).
Proof.
  move => A B Himp.
  have H : (|-- [A](A ->> B) <<->> [A](A ->> -!-!B)). {
    apply: (form_rewrite _ _ (gqm_dnegEquiv B)).
  }
  hAndDestruct H => H _.
  apply: gqm_mp; first by apply H.
  done.
Qed.

Local Example rewrite_test_3 : forall A B, |-- [A](A ->> [A] B) -> |-- [A](A ->> [A]-!-!B).
Proof.
  move => A B Himp.
  have H : (|-- [A](A ->> [A]B) <<->> [A](A ->> [A]-!-!B)). {
    apply: (form_rewrite _ _ (gqm_dnegEquiv B)).
  }
  hAndDestruct H => H _.
  apply: gqm_mp; first by apply H.
  done.
Qed.

Local Example rewrite_test_4 : forall A B, |-- [A](A ->> [A]|[]| B) -> |-- [A](A ->> [A]|[]|-!-!B).
Proof.
  move => A B Himp.
  have H : (|-- [A](A ->> [A]|[]|B) <<->> [A](A ->> [A]|[]|-!-!B)). {
    apply: (form_rewrite _ _ (gqm_dnegEquiv B)).
  }
  hAndDestruct H => H _.
  apply: gqm_mp; first by apply H.
  done.
Qed.

Local Example rewrite_test_5 : forall A B, |-- (A <<->> B) -> |-- (A <<->> -!-!B).
Proof.
  move => A B Himp.
  have H : (|-- (A <<->> B) <<->> (A <<->> -!-!B)). {
    apply (@form_rewrite (fun X => A <<->> X) _ _ _ (gqm_dnegEquiv B)).
  }
  hAndDestruct H => H _.
  apply: gqm_mp; first by apply H.
  done.
Qed.

Local Example rewrite_test_6 : forall A, |-- |A|A -> |-- |A|-!-!A.
Proof.
  move => A Himp.
  have H : (|-- |A|A <<->> |A|-!-!A). {
    apply (@form_rewrite All _ _ _ (gqm_dnegEquiv A)).
  }
  hAndDestruct H => H _.
  apply: gqm_mp; first by apply H.
  done.
Qed.

Lemma tac_rewrite(r: form -> form) `{Rewriteable r} :
  forall {A} {B}, |-- A <<->> B -> |-- r B -> |-- r A.
Proof.
  case: H => Hr.
  move => A B Heq HA.
  hAndDestruct (Hr _ _ Heq) => _ HAB.
  by apply: gqm_mp; first by apply HAB.
Qed.

Lemma gqm_equiv_comm : forall {A} {B}, |-- A <<->> B -> |-- B <<->> A.
Proof.
  move => A B Heq.
  hAndDestruct Heq.
  by hSplit.
Qed.

Lemma tac_rewriteR(r: form -> form) `{Rewriteable r} :
  forall {A} {B}, |-- B <<->> A -> |-- r B -> |-- r A.
Proof.
  move => A B H2 H3.
  apply: (tac_rewrite r (gqm_equiv_comm H2) H3).
Qed.

Local Example rewrite_test_1' : forall A B, |-- A ->> -!-!B -> |-- A ->> B.
Proof.
  move => A B Himp.
  by apply: (tac_rewrite (fun X => A ->> X) (gqm_dnegEquiv B)).
Qed.

Tactic Notation "hRewrite" constr(Heq) constr(fShape) :=
  apply: (tac_rewrite fShape Heq).

Local Example rewrite_test_2' : forall A B, |-- [A](A ->> -!-!B) -> |-- [A](A ->> B).
Proof.
  move => A B Himp.
  by hRewrite (gqm_dnegEquiv B) (fun X => [A](A ->> X)).
Qed.

Local Example rewrite_test_3' : forall A B, |-- [A](A ->> [A] -!-!B) -> |-- [A](A ->> [A]B).
Proof.
  move => A B Himp.
  by hRewrite (gqm_dnegEquiv B) (fun X => [A](A ->> [A] X)).
Qed.

Local Example rewrite_test_4' : forall A B, |-- [A](A ->> [A]|[]| -!-!B) -> |-- [A](A ->> [A]|[]|B).
Proof.
  move => A B Himp.
  by hRewrite (gqm_dnegEquiv B) (fun X => [A](A ->> [A]|[]| X)).
Qed.

Local Example rewrite_test_5' : forall A B, |-- (A <<->> -!-!B) -> |-- (A <<->> B).
Proof.
  move => A B Himp.
  by hRewrite (gqm_dnegEquiv B) (fun X => A <<->> X).
Qed.

Local Example rewrite_test_6' : forall A, |-- |A|-!-!A -> |-- |A|A.
Proof.
  move => A Himp.
  by hRewrite (gqm_dnegEquiv A) (fun X => |A|X).
Qed.